Require Import Metatheory.
Require Import FJ_Definitions.

(********************************************************************)
(* CFJ - AUXILIARIES                                                *)
(********************************************************************)

(* Annotation are modeled as propositional formulas. *)
Notation ann := Prop.

(* Expressions can be annotated if they occur in a list. Note, that
   we cannot safe the annotation directly, for instance at ae_var,
   since such an expression might occur in a context where it cannot
   be annotated; the outermost term of a method or a term in a cast.
   Coq forces us to use the Type instead of Set. *)
Inductive a_exp : Type :=
| ae_var : var -> a_exp
| ae_field : a_exp -> fname -> a_exp
| ae_meth : a_exp -> mname -> list (ann * a_exp) -> a_exp
| ae_new : cname -> list (ann * a_exp) -> a_exp
(* EDIT4 BEGIN *)
| ae_mex : a_exp -> list (ann * a_exp) -> a_exp
(* EDIT4 END *)
.

(* The new environment is a list of "x:T with A", where the variable
   x is bound to the type T and the annotation A. *)
Notation a_env := (list (var * (ann * typ))).

(* Fields and methods can be annotated. Additionally, method
   parameters or parts of the method's term might be annotated. *)
Notation a_flds := (list (fname * (ann * typ))).
Notation a_mths := (list (mname * (ann * (typ * a_env * a_exp)))).

(* In the class table each class can be annotated. In a class tuple
   (C,(AT_C,D,gs,ts,fs,ms)) we store the following information:
   C - class name
   AT_C* - annotation of C
   D - super class of C
   gs* - annotated constructor parameters
   hs* - annotated super call parameters
   ts* - annotated assignments this.f=f;
   fs - annotated fields
   ms - annotated methods
   The elements marked with * were added to store all annotations. *)
Notation a_ctable := (list (cname * (ann * cname * a_flds * a_flds * 
    a_flds * a_flds * a_mths))).

(* As for FJ, we assume a fixed class table for CFJ. As we want to
   reuse the definitions for FJ, we also need to assume a fixed class
   table for a FJ variant. See variant generation on more details. *)
Parameter aCT : a_ctable.

(* The lookup of annotations at the class table for
   - AT: classes
   - ATp: constructor parameters
   - ATs: super call parameters
   - ATa: assignments (this.f:=f;)
   - ATf: fields
   - ATm: methods
   - ATmx: method parameters
   False is returned if it does not exist. A special case is Object,
   for which we always need to return True. *)
Definition AT (C:cname) : ann := (*$\label{annotationlookup}$*)
    match (get C aCT) with
    | None => if C == Object then True else False
    | Some (a, _, _, _, _, _, _) => a
    end.
Definition ATp (C:cname) (f:fname) : ann := 
    match (get C aCT) with
    | None => False
    | Some (_, _, gs, _, _, _, _) =>
        match (get f gs) with
        | None => False
        | Some (a, _) => a
        end 
    end.
Definition ATs (C:cname) (f:fname) : ann := 
    match (get C aCT) with
    | None => False
    | Some (_, _, _, hs, _, _, _) =>
        match (get f hs) with
        | None => False
        | Some (a, _) => a
        end 
    end.
Definition ATa (C:cname) (f:fname) : ann := 
    match (get C aCT) with
    | None => False
    | Some (_, _, _, _, ts, _, _) =>
        match (get f ts) with
        | None => False
        | Some (a, _) => a
        end 
    end.
Definition ATf (C:cname) (f:fname) : ann :=
    match (get C aCT) with
    | None => False
    | Some (_, D, _, _, _, fs, _) =>
        match (get f fs) with
        | None => False
        | Some (a, _) => a
        end 
    end.
Definition ATm (C:cname) (m:mname) : ann :=
    match (get C aCT) with
    | None => False
    | Some (_, _, _, _, _, _, ms) =>
        match (get m ms) with
        | None => False
        | Some (a, _) => a
        end 
    end.

(* Although, the field lookup is not extended in CFJ, we need a new
   definition that uses the CFJ class table. *)
Inductive a_fields : cname -> a_flds -> Prop :=
| a_fields_obj : a_fields Object nil
| a_fields_other : forall C D gs hs ts fs fs' ms, 
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    a_fields D fs' ->
    a_fields C (fs'++fs).
Definition a_field (C:cname) (f:fname) (a:ann) (t:typ) : Prop :=
    exists2 fs, a_fields C fs & binds f (a,t) fs.

(* The lookup of method declarations is extended as follows. A method
   is either defined in all variants where it is used, or the search
   is continued at the super class. *)
Inductive a_method : (*$\label{a_method}$*)
  cname -> mname -> ann -> typ * a_env * a_exp -> Prop :=
| a_method_this : forall (C D:cname) (fs gs hs ts:a_flds) (m:mname)
    (ms:a_mths) (mdecl:typ * a_env * a_exp) (a:ann),
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    binds m (ATm C m,mdecl) ms ->
    (a -> ATm C m) ->
    a_method C m a mdecl
| a_method_notthis : forall (C D:cname) (fs gs hs ts:a_flds) (m:mname)
    (ms:a_mths) (mdecl mdecl':typ * a_env * a_exp) (a:ann),
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    binds m (ATm C m,mdecl') ms ->
    ~(a -> ATm C m) ->
    a_method D m (a /\ ~ ATm C m) mdecl ->
    a_method C m a mdecl
| a_method_super : forall (C D:cname) (fs gs hs ts:a_flds) (m:mname)
    (ms:a_mths) (mdecl:typ * a_env * a_exp) (a:ann),
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT ->
    no_binds m ms ->
    a_method D m a mdecl ->
    a_method C m a mdecl.

(* Evaluation rules are not needed as CFJ product lines are never
   directly evaluated; FJ programs are generated, that can be
   evaluated using the definitions for FJ. *)

(* The subtype relation needs to be defined on the CFJ class table.
   It is the reflexive, transitive closure of the direct subclass
   relation. *)
Definition a_extends (C D : cname) : Prop :=
  exists gs, exists hs, exists ts, exists fs, exists ms,
    binds C (AT C,D,gs,hs,ts,fs,ms) aCT.
Inductive a_sub : typ -> typ -> Prop :=
| a_sub_refl : forall t, a_sub t t
| a_sub_trans : forall t1 t2 t3,
    a_sub t1 t2 -> a_sub t2 t3 -> a_sub t1 t3
| a_sub_extends : forall C D, a_extends C D -> a_sub C D.

Hint Unfold a_field a_extends.
Hint Constructors a_fields a_method a_sub.

(* EDIT4 BEGIN *)
(* The most specific supertype is needed for alternative terms. *)
Inductive mss : typ -> typ -> typ -> Prop :=
| mss_fst : forall C D,
    a_sub D C ->
    mss C D C
| mss_snd : forall C D,
    a_sub C D ->
    mss C D D
| mss_other : forall C D E,
    a_sub C E ->
    a_sub D E ->
    not (exists E', a_sub E' E /\ a_sub C E' /\ a_sub D E') ->
    mss C D E.
Inductive mss' : typ -> list typ -> typ -> Prop :=
| mss_nil : forall (C:typ),
    mss' C nil C
| mss_same : forall (D0 Di C C':typ) (Ds:list typ),
    mss' D0 Ds C ->
    mss Di C C' ->
    mss' D0 (Di::Ds) C'.

(* Simulates a list of annotated types, given a list containing the
   annotations and a list containing the types. *)
Inductive ann_zip : list (ann*a_exp) -> list typ -> a_env
  -> Prop :=
| azip_nil : ann_zip nil nil nil
| azip_cons : forall (a:ann) (t:a_exp) (C:typ) (ts:list (ann*a_exp))
  (Cs:list typ) (Cxs:a_env) (x:var),
    ann_zip ts Cs Cxs ->
    ann_zip ((a,t)::ts) (C::Cs) ((x,(a,C))::Cxs).

Hint Constructors mss mss' ann_zip.
(* EDIT4 END *)

(********************************************************************)
(* CFJ - TYPING                                                     *)
(********************************************************************)

(* As for FJ, term typing is defined mutually. The reachability checks
   are straight forward. *)
Inductive a_typing : (*$\label{a_typing}$*)
  ann -> a_env -> a_exp -> typ -> Prop :=
| a_t_var : forall (a a':ann) (E:a_env) (x:var) (C:typ),
    ok E ->
    binds x (a',C) E ->
    (a -> a') ->
    (a' -> AT C) ->
    a_typing a E (ae_var x) C
| a_t_field : forall (a:ann) (E:a_env) (t0:a_exp) (C0 Ci:typ)
    (fi:fname),
    a_typing a E t0 C0 ->
    a_field C0 fi (ATf C0 fi) Ci ->
    (a -> ATf C0 fi) ->
    (ATf C0 fi -> AT Ci) ->
    a_typing a E (ae_field t0 fi) Ci
| a_t_meth : forall (a:ann) (E:a_env) (Dys:a_env) (t0 t:a_exp)
    (C0 C:typ) (m:mname) (ts:list (ann * a_exp)),
    a_typing a E t0 C0 ->
    a_method C0 m a (C,Dys,t) ->
    a_wide_typings a E ts (imgs Dys) ->
    (a -> AT C) ->
    a_typing a E (ae_meth t0 m ts) C
| a_t_new : forall (a:ann) (E:a_env) (Dfs:a_flds) (C:typ)
    (ts:list (ann * a_exp)),
    a_fields C Dfs ->
    a_wide_typings a E ts (imgs Dfs) ->
    (a -> AT C) ->
    a_typing a E (ae_new C ts) C
(* EDIT4 BEGIN *)
| a_t_mex : forall (a:ann) (E:a_env) (t:a_exp) (ts:list (ann*a_exp))
    (C D0:typ) (Ds:list typ) (Dxs:a_env),
    a_typing a E t D0 ->
    ann_zip ts Ds Dxs ->
    a_wide_typings a E ts (imgs Dxs) ->
    mss' D0 Ds C ->
    (a -> AT C) ->
    a_typing a E (ae_mex t ts) C
(* EDIT4 END *)

with a_wide_typing : (*$\label{a_wide_typing}$*)
  ann -> a_env -> a_exp -> typ -> Prop :=
| a_wt_sub : forall (a:ann) (E:a_env) (e:a_exp) (t t':typ),
    a_typing a E e t -> a_sub t t' -> a_wide_typing a E e t'

with a_wide_typings : (*$\label{a_wide_typings}$*)
  ann -> a_env -> list (ann * a_exp) -> list (ann * typ) -> Prop :=
| a_wts_nil : forall (a:ann) (E:a_env),
    ok E ->
    a_wide_typings a E nil nil
| a_wts_cons : forall (a AT_t AT_Dy:ann) (E:a_env)
    (Ds:list (ann * typ)) (ts:list (ann * a_exp)) (t:a_exp) (C:typ),
    a_wide_typings a E ts Ds ->
    a_wide_typing AT_t E t C ->
    (a -> (AT_t <-> AT_Dy)) ->
    (AT_t -> a) ->
    a_wide_typings a E ((AT_t,t)::ts) ((AT_Dy,C)::Ds).

(* For valid method overriding we need to check, that the annotations
   of the parameter declarations match. *)
Inductive a_can_override_check : (*$\label{a_can_override_check}$*)
  ann -> mname -> cname -> a_env -> cname -> a_env -> Prop :=
| a_coc_nil : forall (a:ann) (m:mname) (C D:cname),
    a_can_override_check a m C nil D nil
| a_coc_cons : forall (a AT_Cixi AT_Diyi:ann) (m:mname) (C D:cname)
    (Cxs Dxs:a_env) (xi yi:var) (Ci:cname),
    a_can_override_check a m C Cxs D Dxs ->
    (a -> (AT_Cixi <-> AT_Diyi)) ->
    a_can_override_check a m C ((xi,(AT_Cixi,Ci))::Cxs)
                             D ((yi,(AT_Diyi,Ci))::Dxs).

(* Checking valid method overriding is straight forward; we simply
   add the reachability checks. *)
Definition a_can_override (*$\label{a_can_override}$*)
  (C D:cname) (m:mname) (C0:typ) (E:a_env) (a:ann) : Prop :=
    forall C0' E' t,
    a_method D m a (C0',E',t) ->
    (*EDIT2 BEGIN*)
    (*a_sub C0 C0' /\*)
    C0 = C0' /\ 
    (*EDIT2 END*)
    a_can_override_check a m C E D E'.

(* The reachability checks in method typing. *)
Inductive a_ok_meth_check : (*$\label{a_ok_meth_check}$*)
  ann -> a_env -> Prop :=
| a_omc_nil : forall (a:ann),
    a_ok_meth_check a nil
| a_omc_cons : forall (a AT_Cixi:ann) (Cxs:a_env) (Ci:typ) (xi:var),
    a_ok_meth_check a Cxs ->
    (AT_Cixi -> AT Ci) ->
    (AT_Cixi -> a) ->
    a_ok_meth_check a ((xi,(AT_Cixi,Ci))::Cxs).

(* Method typing with reachability checks. *)
Definition a_ok_meth (*$\label{a_ok_meth}$*)
  (C D:cname) (m:mname) (C0:typ) (Cxs:a_env) (t0:a_exp) : Prop :=
    let a := ATm C m in
    let E := (this,(AT C,C))::Cxs in
    (*EDIT4 BEGIN*)
    (*(a -> AT C0) /\*)
    (*EDIT4 END*)
    a_ok_meth_check a Cxs /\
    a_can_override C D m C0 Cxs a /\
    a_wide_typing a E t0 C0.
Definition a_ok_meth'
  (C D: cname) (m : mname) (v:ann * (typ * a_env * a_exp)) : Prop :=
    match v with (_,(C0,E,t)) => a_ok_meth C D m C0 E t end.

(* Reachability checks for class typing. *)
Inductive a_ok_class_check : (*$\label{a_ok_class_check}$*)
  ann -> cname -> cname -> a_flds -> a_flds -> a_mths -> Prop :=
| a_occ_nil : forall (a:ann) (C:cname) (D:cname),
    a_ok_class_check a C D nil nil nil
| a_occ_cons_flds1 : forall (a:ann) (C:cname) (D:cname) (fs:a_flds)
    (fi:fname) (Ci:typ),
    a_ok_class_check a C D fs nil nil ->
    (ATf C fi -> a /\ AT Ci) ->
    ((ATf C fi <-> ATa C fi) /\ (ATa C fi <-> ATp C fi)) ->
    a_ok_class_check a C D ((fi,(ATf C fi,Ci))::fs) nil nil
| a_occ_cons_flds2 : forall (a AT_Digi:ann) (C:cname) (D:cname)
    (fs gs:a_flds) (gi:fname) (Di:typ),
    a_ok_class_check a C D fs gs nil ->
    (a -> (ATp C gi <-> AT_Digi)) ->
    (ATp C gi <-> ATs C gi) ->
    (ATp C gi -> a) ->
    a_ok_class_check a C D fs ((gi,(AT_Digi,Di))::gs) nil
| a_occ_cons_mths : forall (a:ann) (C: cname) (D: cname)
    (fs gs:a_flds) (ms:a_mths) (m:mname) (m':typ * a_env * a_exp),
    a_ok_class_check a C D fs gs ms ->
    (ATm C m -> a) ->
    a_ok_class_check a C D fs gs ((m,(ATm C m,m'))::ms).

(* Straight forward extension of class typing. *)
Definition a_ok_class (*$\label{a_ok_class}$*)
  (C:cname) (D:cname) (fs:a_flds) (ms:a_mths) : Prop :=
    let a := AT C in
    (*EDIT1 BEGIN*)
    (*(forall (gs:a_flds),
      fields D gs -> a_ok_class_check a C D fs gs ms /\*)
    (exists gs:a_flds,
      a_fields D gs /\ a_ok_class_check a C D fs gs ms /\
    (*EDIT1 END*)
      ok (gs ++ fs) ) /\
    (a -> AT D) /\ ok ms /\ forall_env (a_ok_meth' C D) ms.
Definition a_ok_class' (C:cname) (v:ann * cname * a_flds * a_flds * 
  a_flds * a_flds * a_mths) : Prop :=
    match v with (_,D,_,_,_,fs,ms) => a_ok_class C D fs ms end.

(* Checks the class table; identical to FJ. *)
Definition a_ok_ctable (*$\label{a_ok_ctable}$*)
  (ct:a_ctable) := ok ct /\ forall_env a_ok_class' ct.

(* A FJ program is well-typed if the class table is well-typed and
   the start term is well-typed. *)
Definition FJ_program (*$\label{FJ_program}$*)
  (t:exp) : Prop :=
    ok_ctable CT /\ wide_typing nil t Object.

(* A CFJ product line is well-typed if the class table is well-typed
   and the start term is well-typed. *)
Definition CFJ_product_line (*$\label{CFJ_product_line}$*)
  (t:a_exp) : Prop :=
    a_ok_ctable aCT /\ a_wide_typing True nil t Object.

Hint Constructors a_typing a_wide_typing a_wide_typings
  a_can_override_check a_ok_meth_check a_ok_class_check.
Hint Unfold a_can_override a_ok_meth a_ok_meth' a_ok_class a_ok_class'
  a_ok_ctable FJ_program CFJ_product_line.

(********************************************************************)
(* VARIANT GENERATION                                               *)
(********************************************************************)

(* We inductively define the variant generation as presented in the
   proposal. As we have a fixed class table for CFJ and a fixed class
   table for a FJ program variant, we also need to assume a fixed
   configuration (feature selection). Nevertheless, both class tables
   and the configuration can be arbitrary.

   As we do not want to formalize the feature model, we simply assume
   it is given as a propositional formula of type Prop. Additionally,
   we force that this formula is included in every annotation a. A
   code fragment annotated with a is removed if a evaluates to false.
   We give an example:
   - variables A B:Prop
   - feature model FM = A -> B
   - annotation a = FM /\ A
   - a evaluates to True, if and only if A and B evaluate to True *)

(* Environments are used for fields and for method parameters. In
   each we need to remove variable assignments not present in the
   current configuration. The annotation is given as we extended
   environments. *)
Inductive variant_env : (*$\label{variant_env}$*)
  a_env -> env -> Prop :=
| ve_nil :
    variant_env nil nil
| ve_cons_y : forall (E:a_env) (E':env) (x:var) (a:ann) (C:typ),
    variant_env E E' ->
    a ->
    variant_env ((x,(a,C))::E) ((x,C)::E')
| ve_cons_n : forall (E:a_env) (E':env) (x:var) (a:ann) (C:typ),
    variant_env E E' ->
    ~a ->
    variant_env ((x,(a,C))::E) E'.

(* Variant generation for terms and lists of terms. A term cannot be
   removed as is, it must be part of a list. For a list of terms the
   annotation is given in a pair together with the term itself. See
   list (ann*a_exp). *)
Inductive variant_term : (*$\label{variant_term}$*)
  a_exp -> exp -> Prop :=
| vt_var : forall (x:var),
    variant_term (ae_var x) (e_var x)
| vt_field : forall (t:a_exp) (t':exp) (f:fname),
    variant_term t t' ->
    variant_term (ae_field t f) (e_field t' f)
| vt_meth : forall (t:a_exp) (t':exp) (m:mname) (ts:list (ann*a_exp))
    (ts':list exp),
    variant_term t t' ->
    variant_terms ts ts' ->
    variant_term (ae_meth t m ts) (e_meth t' m ts')
| vt_new : forall (ts:list (ann*a_exp)) (ts':list exp) (C:typ),
    variant_terms ts ts' ->
    variant_term (ae_new C ts) (e_new C ts')
(* EDIT4 BEGIN *)
| vt_mex_y : forall (t:a_exp) (t':exp) (ts:list (ann*a_exp)) (ts':list exp),
    variant_terms ts (t'::ts') ->
    variant_term (ae_mex t ts) t'
| vt_mex_n : forall (t:a_exp) (t':exp) (ts:list (ann*a_exp)),
    variant_terms ts nil ->
    variant_term t t' ->
    variant_term (ae_mex t ts) t'
(* EDIT4 END *)

with variant_terms : (*$\label{variant_terms}$*)
  list (ann*a_exp) -> list exp -> Prop :=
| vts_nil :
    variant_terms nil nil
| vts_cons_y : forall (ts:list(ann*a_exp)) (ts':list exp) (a:ann)
    (t:a_exp) (t':exp),
    variant_terms ts ts' ->
    a ->
    variant_term t t' ->
    variant_terms ((a,t)::ts) (t'::ts')
| vts_cons_n : forall (ts:list(ann*a_exp)) (ts':list exp) (a:ann)
    (t:a_exp),
    variant_terms ts ts' ->
    ~a ->
    variant_terms ((a,t)::ts) ts'.

(* For mutual inductive definitions we need to generate a useful
   induction principle. *)
Scheme variant_term_ind2 := Minimality for variant_term Sort Prop
with variant_terms_ind2 := Minimality for variant_terms Sort Prop.
Combined Scheme variant_term_mutind from variant_term_ind2,
  variant_terms_ind2.

(* Variant generation for a list of methods. As above, methods are
   only present if Conf -> a, where a is the annotation of the
   method. *)
Inductive variant_methods : (*$\label{variant_methods}$*)
  cname -> a_mths -> mths -> Prop :=
| vms_nil : forall (C:cname),
    variant_methods C nil nil
| vms_cons_y : forall (C:cname) (ms:a_mths) (ms':mths) (m:mname)
    (C0:typ) (Cxs:a_env) (Cxs':env) (t:a_exp) (t':exp),
    variant_methods C ms ms' ->
    variant_term t t' ->
    variant_env Cxs Cxs' ->
    ATm C m ->
    variant_methods C ((m,(ATm C m,(C0,Cxs,t)))::ms)
                      ((m,(C0,Cxs',t'))::ms')
| vms_cons_n : forall (C:cname) (ms:a_mths) (ms':mths) (m:mname)
    (C0:typ) (Cxs:a_env) (t:a_exp),
    variant_methods C ms ms' ->
    ~(ATm C m) ->
    variant_methods C ((m,(ATm C m,(C0,Cxs,t)))::ms) ms'.

(* Variant generation for the class table. Removes classes that are
   not in the current configuration. *)
Inductive variant_ct : (*$\label{variant_ct}$*)
  a_ctable -> ctable -> Prop :=
| vct_nil :
    variant_ct nil nil
| vct_cons_y : forall (ct:a_ctable) (ct':ctable) (C D:cname)
    (fs gs hs ts:a_flds) (fs':flds) (ms:a_mths)(ms':mths),
    variant_ct ct ct' ->
    variant_env fs fs' ->
    variant_methods C ms ms' ->
    AT C ->
    variant_ct ((C,(AT C,D,gs,hs,ts,fs,ms))::ct)
               ((C,(D,fs',ms'))::ct')
| vct_cons_n : forall (ct:a_ctable) (ct':ctable) (C D:cname)
    (fs gs hs ts:a_flds) (fs':flds) (ms:a_mths) (ms':mths),
    variant_ct ct ct' ->
    ~(AT C) ->
    variant_ct ((C,(AT C,D,gs,hs,ts,fs,ms))::ct) ct'.

Hint Constructors variant_term variant_terms variant_env
  variant_methods variant_ct.

(* Almost all following theorems are based on the fact that the FJ
   class table CT is generated from the CFJ class table aCT. Hence,
   we assume this, as a parameter. *)
Parameter aCT_correlates_with_CT : variant_ct aCT CT.

(* Additionally, we often need that the CFJ class table aCT is
   well-typed. We enforce this by a parameter. *)
Parameter aCT_well_typed : a_ok_ctable aCT.

(* The feature model is given as a propositional formula. Like both
   class tables it is modeled using a parameter. We assume that this
   formula is included in every annotation at the class table. *)
Parameter FM : Prop.
